# Generated by Django 2.2.5 on 2019-09-20 05:31

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="CamundaConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "root_url",
                    models.URLField(
                        default="http://localhost:12018/gemeente-utrecht/camunda/",
                        help_text="Root URL where camunda is installed. The REST api path is appended to this.",
                        verbose_name="camunda root",
                    ),
                ),
                (
                    "rest_api_path",
                    models.CharField(
                        default="engine-rest",
                        max_length=255,
                        verbose_name="REST api path",
                    ),
                ),
            ],
            options={"verbose_name": "Camunda configuration"},
        )
    ]



from typing import Dict, List, Union

import inflection


def underscoreize(data: Union[List, Dict, str, None]) -> Union[List, Dict, str, None]:
    if isinstance(data, list):
        return [underscoreize(item) for item in data]

    if isinstance(data, dict):
        new_data = {}
        for key, value in data.items():
            new_key = inflection.underscore(key)
            # variables are dynamic names, can't make assumptions!
            if key == "variables":
                new_data[new_key] = value
            else:
                new_data[new_key] = underscoreize(value)
        return new_data

    return data



"""
Public Python API to interact with Activiti.
"""
from typing import Dict, List, Optional

from .camunda_models import ProcessDefinition, factory
from .client import get_client
from .types import JSONObject


def get_process_definitions() -> List[ProcessDefinition]:
    client = get_client()
    response = client.get("process-definition", {"sortBy": "name", "sortOrder": "asc"})
    return factory(ProcessDefinition, response)


def get_start_form_variables(
    process_key: Optional[str] = None, process_id: Optional[str] = None
) -> Dict[str, JSONObject]:
    """
    Get the start form variables from the Camunda process.

    If defaults are configured in the Camunda process, these will be returned as value.

    A process ID is more specific than a process key, so if both are provided, the
    process ID will be used.
    """
    if not (process_key or process_id):
        raise ValueError("Provide a process key or process ID")

    if process_id:
        endpoint = f"process-definition/{process_id}/form-variables"
    else:
        endpoint = f"process-definition/key/{process_key}/form-variables"

    # TODO: do any necessary type casting
    client = get_client()
    variables = client.get(endpoint, underscoreize=False)

    return variables



import logging
from typing import Dict, Optional, Union

from celery import shared_task

from .client import get_client
from .interface import Variable

logger = logging.getLogger(__name__)


@shared_task
def start_process(
    process_key: Optional[str] = None,
    process_id: Optional[str] = None,
    business_key: Optional[str] = None,
    variables: Dict[str, Union[Variable, dict]] = None,
) -> Dict[str, str]:
    """
    Start a process in the camunda process engine.

    The result is stored in the Celery result backend, so you can match a task
    ID with the resulting process instance.

    :param process_key: the key that a process is deployed with. Will start the
      latest vesion of this process.
    :param process_id: the id of a process. Will start this particular process version.
    :param business_key: the camunda business key that applies.
    :param variables: a dictionary where the keys are variable names relevant
      for the process. Values can be instances of :class:`Variable` for complex/
      custom types, or simple dicts in the camunda format. See
      https://docs.camunda.org/manual/7.11/reference/rest/process-definition/post-start-process-instance/#starting-a-process-instance-at-its-default-initial-activity
      for an example.
    :return: a dict with the details of the started process instance
    """
    logger.debug(
        "Received process start: process_key=%s, process_id=%s", process_key, process_id
    )
    if not (process_key or process_id):
        raise ValueError("Provide a process key or process ID")

    client = get_client()
    variables = variables or {}

    _variables = {
        key: var.serialize() if isinstance(var, Variable) else var
        for key, var in variables.items()
    }

    if process_id:
        endpoint = f"process-definition/{process_id}/start"
    else:
        endpoint = f"process-definition/key/{process_key}/start"

    body = {
        "businessKey": business_key,
        "withVariablesInReturn": False,
        "variables": _variables,
    }

    response = client.post(endpoint, json=body)

    self_rel = next((link for link in response["links"] if link["rel"] == "self"))
    instance_url = self_rel["href"]

    logger.info("Started process instance %s", response["id"])

    return {"instance_id": response["id"], "instance_url": instance_url}



from urllib.parse import urljoin

from django.core.exceptions import ValidationError
from django.db import models
from django.utils.translation import ugettext_lazy as _

from solo.models import SingletonModel


class CamundaConfig(SingletonModel):
    root_url = models.URLField(
        _("camunda root"),
        help_text=_(
            "Root URL where camunda is installed. The REST api "
            "path is appended to this."
        ),
        default="https://camunda.example.com",
    )
    rest_api_path = models.CharField(
        _("REST api path"), max_length=255, default="engine-rest"
    )
    auth_header = models.TextField(
        _("authorization header"),
        blank=True,
        help_text=_(
            "HTTP Authorization header value, required if the API is not open."
        ),
    )

    class Meta:
        verbose_name = _("Camunda configuration")

    def __str__(self):
        return self.api_root

    def clean(self):
        from .client import get_client_class

        client = get_client_class()(config=self)
        try:
            client.request("version")
        except Exception as exc:
            raise ValidationError(
                _(
                    "Invalid API root '{root}', got error {error} while checking the "
                    "version endpoint."
                ).format(root=self.api_root, error=exc)
            )

    def save(self, *args, **kwargs):
        if self.rest_api_path.startswith("/"):
            self.rest_api_path = self.rest_api_path[1:]

        if not self.rest_api_path.endswith("/"):
            self.rest_api_path = f"{self.rest_api_path}/"

        super().save(*args, **kwargs)

    @property
    def api_root(self) -> str:
        assert not self.rest_api_path.startswith("/")
        assert self.rest_api_path.endswith("/")
        return urljoin(self.root_url, self.rest_api_path)


class ProcessInstanceMixin(models.Model):
    # track camunda references
    camunda_process_instance_id = models.CharField(
        _("process instance ID"),
        max_length=100,
        unique=True,
        blank=True,
        null=True,
        default=None,
    )
    camunda_process_instance_url = models.URLField(
        _("process instance URL"), unique=True, blank=True, null=True, default=None
    )

    class Meta:
        abstract = True



# Generated by Django 3.0.3 on 2020-02-27 15:52

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("django_camunda", "0002_camundaconfig_auth_header")]

    operations = [
        migrations.AlterField(
            model_name="camundaconfig",
            name="root_url",
            field=models.URLField(
                default="https://camunda.example.com",
                help_text="Root URL where camunda is installed. The REST api path is appended to this.",
                verbose_name="camunda root",
            ),
        )
    ]



from django.db import models
from django.utils.translation import gettext_lazy as _

from .forms import ProcessDefinitionChoicesField


class ProcessDefinitionField(models.CharField):
    def __init__(self, *args, **kwargs):
        # See https://github.com/camunda/camunda-bpm-platform/blob/master/engine/src/
        # main/resources/org/camunda/bpm/engine/db/create/activiti.mysql.create.engine.sql#L134
        # for FK VARCHAR(64) reference
        kwargs.setdefault("max_length", 64)
        kwargs.setdefault("help_text", _("ID of the process definition in Camunda"))
        super().__init__(*args, **kwargs)

    def formfield(self, **kwargs):
        defaults = {"form_class": ProcessDefinitionChoicesField}
        defaults.update(**kwargs)
        return super().formfield(**defaults)



import base64
from itertools import groupby
from typing import List, Tuple, Union

from django import forms
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from .api import get_process_definitions
from .models import CamundaConfig

Choice = Tuple[str, str]


class CamundaConfigForm(forms.ModelForm):
    basic_auth_username = forms.CharField(
        label=_("Username"),
        required=False,
        help_text=_("Username to authenticate against the Camunda API."),
    )

    basic_auth_password = forms.CharField(
        label=_("Password"),
        required=False,
        help_text=_("Password to authenticate against the Camunda API."),
        widget=forms.PasswordInput,
    )

    class Meta:
        model = CamundaConfig
        fields = (
            "root_url",
            "rest_api_path",
            "basic_auth_username",
            "basic_auth_password",
        )

    def __init__(self, *args, **kwargs):
        config = kwargs.get("instance")
        if config is not None:
            creds = self.set_creds_from_header(config)
            if creds is not None:
                kwargs.setdefault("initial", {})
                kwargs["initial"].update(creds)
        super().__init__(*args, **kwargs)

    def clean(self):
        super().clean()
        self.set_basic_auth_header()

    def set_basic_auth_header(self):
        username = self.cleaned_data.get("basic_auth_username")
        password = self.cleaned_data.get("basic_auth_password")
        if not (username and password):
            return

        basic_auth = f"{username}:{password}"
        b64 = base64.b64encode(basic_auth.encode()).decode()
        self.instance.auth_header = f"Basic {b64}"

    def set_creds_from_header(self, config: CamundaConfig) -> Union[dict, None]:
        if not config.auth_header.startswith("Basic "):
            return

        prefix, b64 = config.auth_header.split(" ", 1)
        decoded = base64.b64decode(b64).decode()
        username, password = decoded.split(":", 1)
        return {"basic_auth_username": username, "basic_auth_password": password}


def get_process_definition_choices() -> List[Tuple[str, List[Choice]]]:
    definitions = get_process_definitions()

    definitions = sorted(definitions, key=lambda d: (d.key, -d.version))
    def_by_key = groupby(definitions, lambda x: x.key)

    choices = [
        (
            format_html(_("Process: <code>{key}</code>"), key=key),
            [
                (
                    definition.id,
                    _("{d.name} (version {d.version})").format(d=definition),
                )
                for definition in definitions
            ],
        )
        for key, definitions in def_by_key
    ]
    return choices


class ProcessDefinitionChoicesField(forms.ChoiceField):
    def __init__(self, *args, **kwargs):
        kwargs.pop("max_length", None)
        kwargs.setdefault("choices", get_process_definition_choices)
        super().__init__(*args, **kwargs)



"""
Middleware to transform data from/to camunda.
"""
import json
from dataclasses import dataclass
from typing import Union


@dataclass
class Variable:
    data: Union[list, dict]

    serialization_data_format = "application/json"
    object_type_name = None

    def serialize(self) -> dict:
        value_info = {
            "serializationDataFormat": self.serialization_data_format,
            "objectTypeName": self.object_type_name,
        }
        self.pre_process()
        return {"type": "json", "value": json.dumps(self.data), "valueInfo": value_info}

    def pre_process(self):
        pass
